/*
  transmit data packets over a radio link as async serial, 8n1

  uses timer1 and the timer interrupt.

  cm@tahina.priv.at 2007
*/

#include "hardware.h"

#include <util/crc16.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/power.h>

#include "radioxmit.h"

static uint8_t xmit_cur_bit;
static uint8_t xmit_cur_byte;
static uint8_t *xmit_cur_ptr;
static uint8_t xmit_count;
static volatile uint8_t xmit_done;

/* send out a packet
   takes the buffer, appends CRC-16, enables transmitter, sends preamble plus
   buffer, disables transmitter.

   packet:

   0xdd 0x16 - preamble
   n data bytes
   2 CRC bytes

*/
void xmit_pkt(uint8_t *buf, uint8_t len) {
  /* timer0: runs at system clock, so we get how many ticks for a
     single bit? */
#define XMIT_BIT_TIME ((F_CPU/XMIT_BAUDRATE) - 1)

  /* prescaler 1 =  CS02:1 = 1
     pwm mode
     COM1A1:COM1A0 = 10 -> clear on compare match, set on start
     OCR1C = XMIT_BIT_TIME
     OCR1A = 0 for 0 bit, XMIT_BIT_TIME for 1 bit
     timer overflow interrupt -> set next bit

     TCCR1 = CTC1 | PWM1A | COM1A1 | CS10
     GTCCR = 0
     TIMSK |= TOIE1
  */
  power_timer1_enable();
  OCR1A = XMIT_BIT_TIME; /* start xmitting "1" */
  TCNT1 = 0;
  TCCR1 = _BV(CTC1) | _BV(PWM1A) | _BV(COM1A1);
  GTCCR = 0;
  TIMSK |= _BV(TOIE1);
  OCR1C = XMIT_BIT_TIME;

  /* set up transmit queue */
  xmit_cur_ptr = buf;
  xmit_cur_byte = *xmit_cur_ptr;
  xmit_count = len;
  xmit_cur_bit = 10;
  xmit_done = 0;

  /* XMIT_ENABLE_WAIT is the startup time from the transmitter's specs;
     we start sending a "1" bit immediately and assume the receiver will
     at least get a part of this to be able to sync to the xmitter or
     whatever */
  XMIT_ON;
  _delay_ms(XMIT_ENABLE_WAIT); 

  /* start timer0 */
  TCCR1 |= _BV(CS10);

  /* now all the magic happens in the ISR */
  
  /* wait for transmission end */
  while(!xmit_done);

  /* send a few bits of "1" at the end */
  _delay_ms(XMIT_END_WAIT);

  XMIT_OFF;
  /* stop timer0: */
  TCCR1 = 0;			/* stop clock */
  power_timer1_disable();	/* disable via power save register */
}

ISR(SIG_OVERFLOW1) {
  if(xmit_done) {
    OCR1A = XMIT_BIT_TIME; /* idle */
  } else {
    if(xmit_cur_bit == 10) {
      OCR1A = 0; /* start bit */
    } else if(xmit_cur_bit == 1) {
      OCR1A = XMIT_BIT_TIME; /* stop bit */
    } else {
      if(xmit_cur_byte & 1)
	OCR1A = XMIT_BIT_TIME;
      else
	OCR1A = 0;
      xmit_cur_byte >>= 1;
    }
    xmit_cur_bit--;
    if(!xmit_cur_bit) {
      xmit_count--;
      if(!xmit_count) {
	xmit_done = 1;
	OCR1A = XMIT_BIT_TIME; /* "1" for idle */
      } else {
	xmit_cur_ptr++;
	xmit_cur_byte = *xmit_cur_ptr;
	xmit_cur_bit = 10;
      }
    }
  }
}
